/*
 * A region mapper for deployment in a new task.
 *
 * Copyright (C) 2022, 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <stdio.h>
#include <string.h>

#include <exec/common.h>
#include <exec/internal_pager.h>
#include <exec/stack.h>
#include <resource/resource_server.h>



/* Initialise a pager for programs, reserving space below the allocated regions
   for the region mapper itself. */

static InternalPager exec_pager(MEM_INTERNAL_PAGER_LIMIT);



int main(int argc, char *argv[])
{
  int debug = (argc > 1) && !strcmp(argv[1], "-v");

  /* Initialise pager regions from region descriptions obtained via the
     auxiliary data. */

  l4re_aux_t *l4re_aux = exec_get_l4re_aux(argc, argv);

  if (debug)
    printf("aux = {%s, %lx, %lx, %lx}\n", l4re_aux->binary, l4re_aux->kip_ds, l4re_aux->dbg_lvl, l4re_aux->ldr_flags);

  /* Skip past the auxiliary structure itself. */

  struct exec_region *region = (struct exec_region *) *((l4_addr_t *) (l4re_aux + 1));

  if (!region)
  {
    printf("Could not find regions.\n");
    return 1;
  }

  for (; region && (region->ds != L4_INVALID_CAP); region++)
  {
    if (debug)
      printf("Adding region: {%lx, %llx, %lx, %lx}\n", region->start, region->start + region->size, region->flags, region->ds);

    MappedRegion r(region->start, region->start + region->size, region->flags, region->ds);
    exec_pager.add(r);
  }

  /* Start the pager. */

  if (debug)
  {
    printf("Initialising pager...\n");
    printf("Pager capability: %lx\n", l4re_env_get_cap(ENV_INTERNAL_PAGER_NAME));
    printf("Main thread: %lx\n", l4re_env()->main_thread);
  }

  /* Initialise the server, enabling notifications. */

  ResourceServer server(&exec_pager);
  long err = server.bind(ENV_INTERNAL_PAGER_NAME);

  if (err)
  {
    printf("Could not bind server: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Starting pager...\n");
  server.start(true);

  printf("Ending pager...\n");
  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
